@@ -111,7 +111,7 @@ public class AsYouTypeFormatter {
   private boolean shouldAddSpaceAfterNationalPrefix = false;
   // This contains the national prefix that has been extracted. It contains only digits without
   // formatting.
-  private String nationalPrefixExtracted = "";
+  private String extractedNationalPrefix = "";
   private StringBuilder nationalNumber = new StringBuilder();
   private List<NumberFormat> possibleFormats = new ArrayList<NumberFormat>();
 
@@ -172,7 +172,7 @@ public class AsYouTypeFormatter {
     return false;
   }
 
-  private void getAvailableFormats(String leadingThreeDigits) {
+  private void getAvailableFormats(String leadingDigits) {
     List<NumberFormat> formatList =
         (isCompleteNumber && currentMetadata.intlNumberFormatSize() > 0)
         ? currentMetadata.intlNumberFormats()
@@ -181,13 +181,14 @@ public class AsYouTypeFormatter {
     for (NumberFormat format : formatList) {
       if (!nationalPrefixIsUsedByCountry || isCompleteNumber ||
           format.isNationalPrefixOptionalWhenFormatting() ||
-          phoneUtil.formattingRuleHasFirstGroupOnly(format.getNationalPrefixFormattingRule())) {
+          PhoneNumberUtil.formattingRuleHasFirstGroupOnly(
+              format.getNationalPrefixFormattingRule())) {
         if (isFormatEligible(format.getFormat())) {
           possibleFormats.add(format);
         }
       }
     }
-    narrowDownPossibleFormats(leadingThreeDigits);
+    narrowDownPossibleFormats(leadingDigits);
   }
 
   private boolean isFormatEligible(String format) {
@@ -199,16 +200,18 @@ public class AsYouTypeFormatter {
     Iterator<NumberFormat> it = possibleFormats.iterator();
     while (it.hasNext()) {
       NumberFormat format = it.next();
-      if (format.leadingDigitsPatternSize() > indexOfLeadingDigitsPattern) {
-        Pattern leadingDigitsPattern =
-            regexCache.getPatternForRegex(
-                format.getLeadingDigitsPattern(indexOfLeadingDigitsPattern));
-        Matcher m = leadingDigitsPattern.matcher(leadingDigits);
-        if (!m.lookingAt()) {
-          it.remove();
-        }
-      } // else the particular format has no more specific leadingDigitsPattern, and it should be
-        // retained.
+      if (format.leadingDigitsPatternSize() == 0) {
+        // Keep everything that isn't restricted by leading digits.
+        continue;
+      }
+      int lastLeadingDigitsPattern =
+          Math.min(indexOfLeadingDigitsPattern, format.leadingDigitsPatternSize() - 1);
+      Pattern leadingDigitsPattern = regexCache.getPatternForRegex(
+          format.getLeadingDigitsPattern(lastLeadingDigitsPattern));
+      Matcher m = leadingDigitsPattern.matcher(leadingDigits);
+      if (!m.lookingAt()) {
+        it.remove();
+      }
     }
   }
 
@@ -267,7 +270,7 @@ public class AsYouTypeFormatter {
     lastMatchPosition = 0;
     currentFormattingPattern = "";
     prefixBeforeNationalNumber.setLength(0);
-    nationalPrefixExtracted = "";
+    extractedNationalPrefix = "";
     nationalNumber.setLength(0);
     ableToFormat = true;
     inputHasFormatting = false;
@@ -352,7 +355,7 @@ public class AsYouTypeFormatter {
         if (attemptToExtractIdd()) {
           isExpectingCountryCallingCode = true;
         } else {  // No IDD or plus sign is found, might be entering in national format.
-          nationalPrefixExtracted = removeNationalPrefixFromNationalNumber();
+          extractedNationalPrefix = removeNationalPrefixFromNationalNumber();
           return attemptToChooseFormattingPattern();
         }
       default:
@@ -362,7 +365,7 @@ public class AsYouTypeFormatter {
           }
           return prefixBeforeNationalNumber + nationalNumber.toString();
         }
-        if (possibleFormats.size() > 0) {  // The formatting pattern is already chosen.
+        if (possibleFormats.size() > 0) {  // The formatting patterns are already chosen.
           String tempNationalNumber = inputDigitHelper(nextChar);
           // See if the accrued digits can be formatted properly already. If not, use the results
           // from inputDigitHelper, which does formatting based on the formatting pattern chosen.
@@ -390,19 +393,24 @@ public class AsYouTypeFormatter {
     return attemptToChooseFormattingPattern();
   }
 
+  // @VisibleForTesting
+  String getExtractedNationalPrefix() {
+    return extractedNationalPrefix;
+  }
+
   // Some national prefixes are a substring of others. If extracting the shorter NDD doesn't result
   // in a number we can format, we try to see if we can extract a longer version here.
   private boolean ableToExtractLongerNdd() {
-    if (nationalPrefixExtracted.length() > 0) {
+    if (extractedNationalPrefix.length() > 0) {
       // Put the extracted NDD back to the national number before attempting to extract a new NDD.
-      nationalNumber.insert(0, nationalPrefixExtracted);
+      nationalNumber.insert(0, extractedNationalPrefix);
       // Remove the previously extracted NDD from prefixBeforeNationalNumber. We cannot simply set
       // it to empty string because people sometimes incorrectly enter national prefix after the
       // country code, e.g. +44 (0)20-1234-5678.
-      int indexOfPreviousNdd = prefixBeforeNationalNumber.lastIndexOf(nationalPrefixExtracted);
+      int indexOfPreviousNdd = prefixBeforeNationalNumber.lastIndexOf(extractedNationalPrefix);
       prefixBeforeNationalNumber.setLength(indexOfPreviousNdd);
     }
-    return !nationalPrefixExtracted.equals(removeNationalPrefixFromNationalNumber());
+    return !extractedNationalPrefix.equals(removeNationalPrefixFromNationalNumber());
   }
 
   private boolean isDigitOrLeadingPlusSign(char nextChar) {
@@ -476,7 +484,8 @@ public class AsYouTypeFormatter {
     // We start to attempt to format only when at least MIN_LEADING_DIGITS_LENGTH digits of national
     // number (excluding national prefix) have been entered.
     if (nationalNumber.length() >= MIN_LEADING_DIGITS_LENGTH) {
-      getAvailableFormats(nationalNumber.substring(0, MIN_LEADING_DIGITS_LENGTH));
+
+      getAvailableFormats(nationalNumber.toString());
       // See if the accrued digits can be formatted properly already.
       String formattedNumber = attemptToFormatAccruedDigits();
       if (formattedNumber.length() > 0) {
@@ -599,6 +608,9 @@ public class AsYouTypeFormatter {
     }
     String countryCodeString = Integer.toString(countryCode);
     prefixBeforeNationalNumber.append(countryCodeString).append(SEPARATOR_BEFORE_NATIONAL_NUMBER);
+    // When we have successfully extracted the IDD, the previously extracted NDD should be cleared
+    // because it is no longer valid.
+    extractedNationalPrefix = "";
     return true;
   }
 
